@@ -26,7 +26,7 @@ from cdd.docstring_utils import (
     Style,
     derive_docstring_format,
 )
-from cdd.emitter_utils import interpolate_defaults
+from cdd.emit.emitter_utils import interpolate_defaults
 from cdd.pure_utils import (
     code_quoted,
     count_iter_items,
@@ -216,7 +216,11 @@ def _scan_phase_numpydoc_and_google(
 
     # Scan all lines so that that each element in `stacker` refers to one 'unit'
     stacker, docstring_lines, line = [], docstring.splitlines(), None
-    first_indent = count_iter_items(takewhile(str.isspace, docstring_lines[0]))
+    first_indent = (
+        count_iter_items(takewhile(str.isspace, docstring_lines[0]))
+        if docstring_lines
+        else 0
+    )
     for line_no, line in enumerate(docstring_lines):
         indent = count_iter_items(takewhile(str.isspace, line))
 
@@ -237,7 +241,9 @@ def _scan_phase_numpydoc_and_google(
                     takewhile(
                         partial(le, return_indent),
                         map(
-                            lambda l: count_iter_items(takewhile(str.isspace, l)),
+                            lambda line_: count_iter_items(
+                                takewhile(str.isspace, line_)
+                            ),
                             docstring_lines[line_no + 3 :],
                         ),
                     )
@@ -552,7 +558,7 @@ def _infer_default(_param, infer_type):
     :type infer_type: ```bool```
     """
     if isinstance(
-        _param["default"], (ast.Str, ast.Num, ast.Constant, ast.NameConstant)
+        _param["default"], (ast.Str, ast.Bytes, ast.Num, ast.Constant, ast.NameConstant)
     ):
         _param["default"] = get_value(_param["default"])
     if _param.get("default", False) in none_types:
@@ -844,11 +850,12 @@ def _fill_doc_with_afterward(scanned):
     ):
         del scanned["scanned_afterward"]
     else:
-        scanned["doc"] += (
-            "\n"
+        scanned["doc"] += "{maybe_nl}{scanned_afterward_str}".format(
+            maybe_nl="\n"
             if scanned["scanned_afterward"] and scanned["scanned_afterward"][0] == ""
-            else ""
-        ) + "\n".join(scanned["scanned_afterward"])
+            else "",
+            scanned_afterward_str="\n".join(scanned["scanned_afterward"]),
+        )
 
 
 def _parse_phase_rest(
